/*
 * Copyright (C) 2000-2002 HighPoint Technologies, Inc.
 * HPT3xx UDMA/ATA RAID Controller Device Driver for Linux
 * File:	hptglb.h
 */
#ifndef _HPTGLB_H_
#define _HPTGLB_H_

#pragma pack(1)

typedef long LONG, *PLONG;
typedef unsigned long ULONG, *PULONG;
typedef unsigned long DWORD, *PDWORD, BOOL;
typedef unsigned int  UINT, *PUINT;
typedef unsigned char BYTE, *PBYTE;
typedef unsigned char UCHAR, *PUCHAR;
typedef UCHAR BOOLEAN;
typedef unsigned short USHORT, *PUSHORT;
typedef unsigned short WORD, *PWORD;
typedef void VOID, *PVOID;
typedef UCHAR * ADDRESS;

/*
 * driver core
 */
#ifndef SUPPORT_SATA
#define USE_MULTIPLE           // Support Read/Write Multiple command
#endif

#define USE_DMA             /* Support DMA operation */
#define SUPPORT_48BIT_LBA	/* Support 48 bit LBA */

/*
 * IDE status definitions
 */
#define IDE_STATUS_ERROR             0x01 /* Error Occurred in Execution    */
#define IDE_STATUS_INDEX             0x02	/* is vendor specific             */
#define IDE_STATUS_CORRECTED_ERROR   0x04	/* Corrected Data                 */
#define IDE_STATUS_DRQ               0x08	/* Ready to transfer data         */
#define IDE_STATUS_DSC               0x10	/* not defined in ATA-2           */
#define IDE_STATUS_DWF               0x20	/* Device Fault has been detected */
#define IDE_STATUS_DRDY              0x40	/* Device Ready to accept command */
#define IDE_STATUS_IDLE              0x50	/* Device is OK                   */
#define IDE_STATUS_BUSY              0x80	/* Device Busy, must wait         */

#define IDE_ERROR_BAD_BLOCK          0x80 /* Reserved now                   */
#define IDE_ERROR_DATA_ERROR         0x40	/* Uncorreectable  Data Error     */
#define IDE_ERROR_MEDIA_CHANGE       0x20	/* Media Changed                  */
#define IDE_ERROR_ID_NOT_FOUND       0x10	/* ID Not Found                   */
#define IDE_ERROR_MEDIA_CHANGE_REQ   0x08	/* Media Change Requested         */
#define IDE_ERROR_COMMAND_ABORTED    0x04	/* Aborted Command                */
#define IDE_ERROR_TRACK0_NOT_FOUND   0x02	/* Track 0 Not Found              */
#define IDE_ERROR_ADDRESS_NOT_FOUND  0x01	/* Address Mark Not Found         */
/*
 * IDE command definitions
 */
#define IDE_COMMAND_RECALIBRATE      0x10 /* Recalibrate                    */
#define IDE_COMMAND_READ             0x20	/* Read Sectors with retry        */
#define IDE_COMMAND_WRITE            0x30	/* Write Sectors with retry       */
#define IDE_COMMAND_VERIFY           0x40	/* Read Verify Sectors with Retry */
#define IDE_COMMAND_SEEK             0x70	/* Seek                           */
#define IDE_COMMAND_SET_DRIVE_PARAMETER   0x91 /* Initialize Device Parmeters */
#define IDE_COMMAND_GET_MEDIA_STATUS 0xDA
#define IDE_COMMAND_DOOR_LOCK        0xDE	/* Door Lock                      */
#define IDE_COMMAND_DOOR_UNLOCK      0xDF	/* Door Unlock 						 */
#define IDE_COMMAND_ENABLE_MEDIA_STATUS   0xEF /* Set Features              */
#define IDE_COMMAND_IDENTIFY         0xEC /* Identify Device                */
#define IDE_COMMAND_MEDIA_EJECT      0xED
#define IDE_COMMAND_SET_FEATURES     0xEF /* IDE set features command       */

#define IDE_COMMAND_FLUSH_CACHE      0xE7
#define IDE_COMMAND_STANDBY_IMMEDIATE 0xE0

#ifdef USE_MULTIPLE
#define IDE_COMMAND_READ_MULTIPLE    0xC4	/* Read Multiple                  */
#define IDE_COMMAND_WRITE_MULTIPLE   0xC5	/* Write Multiple                 */
#define IDE_COMMAND_SET_MULTIPLE     0xC6	/* Set Multiple Mode              */
#endif //USE_MULTIPLE

#ifdef USE_DMA
#define IDE_COMMAND_DMA_READ        0xc9  /* IDE DMA read command           */
#define IDE_COMMAND_DMA_WRITE       0xcb  /* IDE DMA write command          */
#endif //USE_DMA
/*
 * IDE_COMMAND_SET_FEATURES
 */
#define FT_USE_ULTRA        0x40    /* Set feature for Ultra DMA           */
#define FT_USE_MWDMA        0x20    /* Set feature for MW DMA					*/
#define FT_USE_SWDMA        0x10    /* Set feature for SW DMA					*/
#define FT_USE_PIO          0x8     /* Set feature for PIO						*/
#define FT_DISABLE_IORDY    0x10    /* Set feature for disabling IORDY		*/
/*
 * IDE drive control definitions
 */
#define IDE_DC_DISABLE_INTERRUPTS    0x02
#define IDE_DC_RESET_CONTROLLER      0x04
#define IDE_DC_REENABLE_CONTROLLER   0x00
/*
 *	Atapi Error Status
 */
#define IDE_ERROR_END_OF_MEDIA       IDE_ERROR_TRACK0_NOT_FOUND 
#define IDE_ERROR_ILLEGAL_LENGTH     IDE_ERROR_ADDRESS_NOT_FOUND

#define MODE_DSP_WRITE_PROTECT  0x80

#ifdef SUPPORT_48BIT_LBA
#define IDE_COMMAND_READ_DMA_EXT       0x25
#define IDE_COMMAND_READ_QUEUE_EXT		0x26
#define IDE_COMMAND_READ_MULTIPLE_EXT	0x29
#define IDE_COMMAND_READ_MAX_ADDR		0x27
#define IDE_COMMAND_READ_EXT				0x24
#define IDE_COMMAND_VERIFY_EXT			0x42
#define IDE_COMMAND_SET_MULTIPLE_EXT	0x37
#define IDE_COMMAND_WRITE_DMA_EXT		0x35
#define IDE_COMMAND_WRITE_QUEUE_EXT		0x36
#define IDE_COMMAND_WRITE_EXT				0x34
#define IDE_COMMAND_WRITE_MULTIPLE_EXT	0x39
#endif // SUPPORT_48BIT_LBA

/*
 * IDENTIFY data
 */
typedef struct _IDENTIFY_DATA {
    USHORT GeneralConfiguration;            // 00 00
    USHORT NumberOfCylinders;               // 02  1
    USHORT Reserved1;                       // 04  2
    USHORT NumberOfHeads;                   // 06  3
    USHORT UnformattedBytesPerTrack;        // 08  4
    USHORT UnformattedBytesPerSector;       // 0A  5
    USHORT SectorsPerTrack;                 // 0C  6
    USHORT VendorUnique1[3];                // 0E  7-9
    USHORT SerialNumber[10];                // 14  10-19
    USHORT BufferType;                      // 28  20
    USHORT BufferSectorSize;                // 2A  21
    USHORT NumberOfEccBytes;                // 2C  22
    USHORT FirmwareRevision[4];             // 2E  23-26
    USHORT ModelNumber[20];                 // 36  27-46
    UCHAR  MaximumBlockTransfer;            // 5E  47
    UCHAR  VendorUnique2;                   // 5F
    USHORT DoubleWordIo;                    // 60  48
    USHORT Capabilities;                    // 62  49
    USHORT Reserved2;                       // 64  50
    UCHAR  VendorUnique3;                   // 66  51
    UCHAR  PioCycleTimingMode;              // 67
    UCHAR  VendorUnique4;                   // 68  52
    UCHAR  DmaCycleTimingMode;              // 69
    USHORT TranslationFieldsValid;          // 6A  53
    USHORT NumberOfCurrentCylinders;        // 6C  54
    USHORT NumberOfCurrentHeads;            // 6E  55
    USHORT CurrentSectorsPerTrack;          // 70  56
    ULONG  CurrentSectorCapacity;           // 72  57-58
    USHORT CurrentMultiSectorSetting;       // 76  59
    ULONG  UserAddressableSectors;          // 78  60-61
    UCHAR  SingleWordDMASupport;            // 7C  62
    UCHAR  SingleWordDMAActive;             // 7D
    UCHAR  MultiWordDMASupport;         	// 7E  63
    UCHAR  MultiWordDMAActive;              // 7F
    UCHAR  AdvancedPIOModes;                // 80  64
    UCHAR  Reserved4;                       // 81
    USHORT MinimumMWXferCycleTime;          // 82  65
    USHORT RecommendedMWXferCycleTime;      // 84  66
    USHORT MinimumPIOCycleTime;             // 86  67
    USHORT MinimumPIOCycleTimeIORDY;        // 88  68
    USHORT Reserved5[2];                    // 8A  69-70
    USHORT ReleaseTimeOverlapped;           // 8E  71
    USHORT ReleaseTimeServiceCommand;       // 90  72
    USHORT MajorRevision;                   // 92  73
    USHORT MinorRevision;                   // 94  74
    USHORT MaxQueueDepth;                   // 96  75
	USHORT SataCapability;                  //     76
    USHORT Reserved6[9];                    // 98   77-85
    USHORT CommandSupport;                  //     86
    USHORT CommandEnable;                   //     87
    USHORT UtralDmaMode;                    /*     88 */
    USHORT Reserved7[11];                   //     89-99
    ULONG  Lba48BitLow;						//     101-100
    ULONG  Lba48BitHigh;					//     103-102
    USHORT Reserved8[23];                   //     104-126
    USHORT SpecialFunctionsEnabled;         //     127
    USHORT Reserved9[128];                  //     128-255

} IDENTIFY_DATA, *PIDENTIFY_DATA;

typedef struct _CONFIGURATION_IDENTIFY_DATA {
	USHORT Revision;
	USHORT MWDMAModeSupported;
	USHORT UDMAModeSupported;
	ULONG  MaximumLbaLow;
	ULONG  MaximumLbaHigh;
	USHORT CommandSupport;
	USHORT Reserved[247];
	UCHAR  Signature; /* 0xA5 */
	UCHAR  CheckSum;
}
CONFIGURATION_IDENTIFY_DATA, *PCONFIGURATION_IDENTIFY_DATA;

/*
 * Define Mode parameter header.
 */
typedef struct _MODE_PARAMETER_HEADER {
	UCHAR ModeDataLength;
	UCHAR MediumType;
	UCHAR DeviceSpecificParameter;
	UCHAR BlockDescriptorLength;
}MODE_PARAMETER_HEADER, *PMODE_PARAMETER_HEADER;
/*
 * SCSI bus status codes.
 */
#define SCSISTAT_GOOD                  0x00
#define SCSISTAT_CHECK_CONDITION       0x02
#define SCSISTAT_CONDITION_MET         0x04
#define SCSISTAT_BUSY                  0x08
#define SCSISTAT_INTERMEDIATE          0x10
#define SCSISTAT_INTERMEDIATE_COND_MET 0x14
#define SCSISTAT_RESERVATION_CONFLICT  0x18
#define SCSISTAT_COMMAND_TERMINATED    0x22
#define SCSISTAT_QUEUE_FULL            0x28

typedef struct _SENSE_DATA {
	UCHAR ErrorCode:7;
	 UCHAR Valid:1;
	UCHAR SegmentNumber;
	UCHAR SenseKey:4;
	UCHAR Reserved:1;
	UCHAR IncorrectLength:1;
	UCHAR EndOfMedia:1;
	UCHAR FileMark:1;
	UCHAR Information[4];
	UCHAR AdditionalSenseLength;
	UCHAR CommandSpecificInformation[4];
	UCHAR AdditionalSenseCode;
	UCHAR AdditionalSenseCodeQualifier;
	UCHAR FieldReplaceableUnitCode;
	UCHAR SenseKeySpecific[3];
} SENSE_DATA, *PSENSE_DATA;

/*
 * Sense codes
 */
#define SCSI_SENSE_NO_SENSE         0x00
#define SCSI_SENSE_RECOVERED_ERROR  0x01
#define SCSI_SENSE_NOT_READY        0x02
#define SCSI_SENSE_MEDIUM_ERROR     0x03
#define SCSI_SENSE_HARDWARE_ERROR   0x04
#define SCSI_SENSE_ILLEGAL_REQUEST  0x05
#define SCSI_SENSE_UNIT_ATTENTION   0x06
#define SCSI_SENSE_DATA_PROTECT     0x07
#define SCSI_SENSE_BLANK_CHECK      0x08
#define SCSI_SENSE_UNIQUE           0x09
#define SCSI_SENSE_COPY_ABORTED     0x0A
#define SCSI_SENSE_ABORTED_COMMAND  0x0B
#define SCSI_SENSE_EQUAL            0x0C
#define SCSI_SENSE_VOL_OVERFLOW     0x0D
#define SCSI_SENSE_MISCOMPARE       0x0E
#define SCSI_SENSE_RESERVED         0x0F
/*
 * Additional Sense codes
 */
#define SCSI_ADSENSE_NO_SENSE       0x00
#define SCSI_ADSENSE_MAN_INTERV     0x03
#define SCSI_ADSENSE_LUN_NOT_READY  0x04
#define SCSI_ADSENSE_ILLEGAL_COMMAND 0x20
#define SCSI_ADSENSE_ILLEGAL_BLOCK  0x21
#define SCSI_ADSENSE_INVALID_LUN    0x25
#define SCSI_ADSENSE_SELECT_TIMEOUT 0x45
#define SCSI_ADSENSE_MUSIC_AREA     0xA0
#define SCSI_ADSENSE_DATA_AREA      0xA1
#define SCSI_ADSENSE_VOLUME_OVERFLOW 0xA7

#define SCSI_ADSENSE_NO_MEDIA_IN_DEVICE 0x3a
#define SCSI_ADWRITE_PROTECT        0x27
#define SCSI_ADSENSE_MEDIUM_CHANGED 0x28
#define SCSI_ADSENSE_BUS_RESET      0x29
#define SCSI_ADSENSE_TRACK_ERROR    0x14
#define SCSI_ADSENSE_SEEK_ERROR     0x15
#define SCSI_ADSENSE_REC_DATA_NOECC 0x17
#define SCSI_ADSENSE_REC_DATA_ECC   0x18
#define SCSI_ADSENSE_ILLEGAL_MODE   0x64
#define SCSI_ADSENSE_BAD_CDB        0x24
#define SCSI_ADSENSE_BAD_PARM_LIST  0x26
#define SCSI_ADSENSE_CANNOT_READ_MEDIUM  0x30

typedef struct _INQUIRYDATA {
	UCHAR DeviceType : 5;
	UCHAR DeviceTypeQualifier : 3;
	UCHAR DeviceTypeModifier : 7;
	UCHAR RemovableMedia : 1;
	UCHAR Versions;
	UCHAR ResponseDataFormat;
	UCHAR AdditionalLength;
	UCHAR Reserved[2];
	UCHAR SoftReset : 1;
	UCHAR CommandQueue : 1;
	UCHAR Reserved2 : 1;
	UCHAR LinkedCommands : 1;
	UCHAR Synchronous : 1;
	UCHAR Wide16Bit : 1;
	UCHAR Wide32Bit : 1;
	UCHAR RelativeAddressing : 1;
	UCHAR VendorId[8];
	UCHAR ProductId[16];
	UCHAR ProductRevisionLevel[4];
	UCHAR VendorSpecific[20];
	UCHAR Reserved3[40];
} INQUIRYDATA, *PINQUIRYDATA;

// Inquiry defines. Used to interpret data returned from target as result
// of inquiry command.
//
// DeviceType field
//
#define DIRECT_ACCESS_DEVICE            0x00    // disks
#define SEQUENTIAL_ACCESS_DEVICE        0x01    // tapes
#define PRINTER_DEVICE                  0x02
#define PROCESSOR_DEVICE                0x03
#define WRITE_ONCE_READ_MULTIPLE_DEVICE 0x04
#define READ_ONLY_DIRECT_ACCESS_DEVICE  0x05    // CD_ROM
#define SCANNER_DEVICE                  0x06
#define LOGICAL_UNIT_NOT_PRESENT_DEVICE 0x7F

/*
 * Read Capacity Data - returned in Big Endian format
 */
typedef struct _READ_CAPACITY_DATA {
	ULONG LogicalBlockAddress;
	ULONG BytesPerBlock;
} READ_CAPACITY_DATA, *PREAD_CAPACITY_DATA;
/*
 * Bus Master Interface
 */
#define BMI_CMD             0       // Bus master IDE command register offset
#define BMI_STS             2       // Bus master IDE status register offset
#define BMI_DTP             4       // Bus master IDE descriptor table
									// pointer register offset

#define BMI_CMD_STARTREAD   9       // Start write (read from disk)
#define BMI_CMD_STARTWRITE  1       // Start read (write to disk)
#define BMI_CMD_STOP        0       // Stop BM DMI

#define BMI_STS_ACTIVE      1       // RO:   bus master IDE active
#define BMI_STS_ERROR       2       // R/WC: IDE dma error
#define BMI_STS_INTR        4       // R/WC: interrupt happen
#define BMI_STS_DRV0EN      0x20    // R/W:  drive0 is capable of DMA xfer
#define BMI_STS_DRV1EN      0x40    // R/W:  drive1 is capable of DMA xfer

/* excluded_flags */
#define EXCLUDE_HPT366    0
#define EXCLUDE_BUFFER    15
#define EXCLUDE_OSREQ_START 16

#define HA_NUM_START  1
#define BUS_NUM_START 1
#define ID_NUM_START  1

#define DEFAULT_TIMING    0	 // Use Mode-0 as default timing


/*
 * event types
 */
#define ET_DEVICE_REMOVED	1	/* device removed */
#define ET_DEVICE_PLUGGED	2	/* device plugged */
#define ET_DEVICE_ERROR		3	/* device I/O error */
#define ET_REBUILD_STARTED  4
#define ET_REBUILD_ABORTED  5
#define ET_REBUILD_FINISHED 6
#define ET_SPARE_TOOK_OVER  7
#define ET_REBUILD_FAILED	8
/* SrbFlags */
#define ARRAY_FORCE_PIO   1

/*
 * Description: Bad Device
 */
typedef struct _BadModeList {
	UCHAR      UltraDMAMode;	 /* 0xFF not support, others useable mode */
	UCHAR      DMAMode;			 /* 0xFF not support, others useable mode */
	UCHAR      PIOMode;			 /* 0xFF not support, others useable mode */
	UCHAR      length;			 /* length of identify string */
	char       *name;			 /* point to identify string */
} BadModeList, *PBadModeList;

#define HPT366_ONLY   0x20		 /* only modify if it is the HPT366 */
#define HPT368_ONLY   0x40		 /* only modify if it is the HPT368 */
#define HPT370_ONLY   0x80		 /* only modify if it is the HPT370 */

/*
 * The time when the error happend
 */
typedef struct _Time_Record {
   UINT    seconds:6;      /* 0 - 59 */
   UINT    minutes:6;      /* 0 - 59 */
   UINT    month:4;        /* 1 - 12 */
   UINT    hours:6;        /* 0 - 59 */
   UINT    day:5;          /* 1 - 31 */
   UINT    year:5;         /* 0=2000, 31=2031 */
} TimeRecord;

/*
 * Error Record
 */
typedef struct _ErrorLog {
   TimeRecord  Time;           /* the time when the error happend */
   UCHAR       ErrorReason;    /* Error Reason or command, see below */
   UCHAR       DiskSeq;        /* Which disk has the error */
   UCHAR       AtaStatus;      /* Ata error status (read from 1x7) */
   UCHAR       nSectors;       /* the number of sectors for I/O command */
   LONG        LBA;            /* the LBA of error location on the disk */
} ErrorLog, *PErrorLog;
#define MAX_ERROR_LOG  32      /* Maximum error log in the info block */

/*
 * Array information block on disk (LBA == RECODR_LBA)
 */
typedef union _ArrayBlock {
	struct{
		UCHAR		unused[30];     /* to be compatible with RAID BIOS */
		WORD      	old_flag;        /* see above */

		ULONG		Signature;      /* 0 This block is vaild array info block */
		ULONG		StripeStamp;    /* 4 Group mark to link disks into array */
		ULONG		MirrorStamp;    /* 8 RAID 0+1 mark to link two RAID0 */
		ULONG		order;          /*12 Mirror sequence. 0 - source, 1 - copy */

		UCHAR		nDisks;         /*16 How many of disks in the array */
		UCHAR		BlockSizeShift; /*17 Block size == (1 << BlockSizeShift) */
		UCHAR		ArrayType;      /*18 See below */
		UCHAR		DeviceNum;      /*19 The sequence number of this disk in the array */

		ULONG		capacity;       /*20 the capacity of the array */

		ULONG		DeviceModeSelect;/*24 disk data transfer mode, 0-4 PIO0-4 */
								    /* 5-7 MW DMA0-3, 8-13 UDMA0-5, 0xFF default */
		ULONG		ModeBootSig;    /*28 The BootDisk is vaild or ont */
		UCHAR		BootDisk;       /*32 AF_BOOT_DISK this disk was selected as */
								    /*   boot disk. others is not */

		UCHAR		ProtectBootSector;/*33 AF_PROTECT_BS protect boot secter from writing */
		UCHAR		nErrorLog;      /*34 the number of error log in errorlog */
		UCHAR		ErrorLogStart;  /*35 the index of start one */
		ErrorLog	errorLog[MAX_ERROR_LOG]; /* 36 error log records */

		ULONG		reserved0;
		ULONG		reserved1;
		ULONG		reserved2;
		ULONG		reserved3;

		ULONG		RebuiltSector;	/* number of sectors that have been rebuilt */
		UCHAR		Validity;		/* The usable flag of the device, see below */
		UCHAR		ArrayName[32];	/* The Name of the array */	//added by wx 12/25/00

	} data;
	UCHAR		reserved[512];	  // to keep the arrayblock 512 bytes
} ArrayBlock;

#define LOC_ARRAY_BLK   ArrayBlock ArrayBlk
#define ARG_ARRAY_BLK ,(PUSHORT)&ArrayBlk

#define RECODR_LBA     9       /* the LBA of the information block on disk */

/* Signature */
#define HPT_ARRAY_OLD  0x5a7816fc /* this is old vaild array block */
#define HPT_ARRAY_NEW  0x5a7816f0 /* this is	new vaild array block */
#define HPT_TMP_SINGLE 0x5a7816fd /* the array including this disk is broken */
#define HPT_CHK_BOOT   0x12345678 /* user set the boot mark */
#define HPT_MODE_SET   0x5a874600 /* user set the data transfer mode for it */
#define DEVICE_MODE_SET(x) ((x & 0xFFFFFF00)==HPT_MODE_SET)

#define MAX_MEMBERS       7    // Maximum members in an array
#define MAX_ADAPTERS      4    // Max Board we can supported in system
#define MAX_DEVICES       (MAX_ADAPTERS * 4)
#define MAX_V_DEVICE      MAX_DEVICES

#define TRUE             1
#define FALSE            0

#define INLINE inline
#define IN
#define OUT
#define OPTIONAL

/*
 * Undef marcos which may conflict.
 */
#undef START_STOP

/*
 * IDE IO Port definition
 */
typedef struct _IDE_REGISTERS_1 {
	USHORT Data;               /* RW: Data port                  */
	UCHAR BlockCount;			 /* RW: Sector count               */
	UCHAR BlockNumber;			 /* RW: Sector number & LBA 0-7    */
	UCHAR CylinderLow;			 /* RW: Cylinder low & LBA 8-15    */
	UCHAR CylinderHigh;			 /* RW: Cylinder hign & LBA 16-23  */
	UCHAR DriveSelect;			 /* RW: Drive/head & LBA 24-27     */
	UCHAR Command;				 /* RO: Status WR:Command          */
} IDE_REGISTERS_1, *PIDE_REGISTERS_1;


/*
 * IDE Control Register File 
 */

typedef struct _IDE_REGISTERS_2 {
	UCHAR AlternateStatus;     /* RW: device control port        */
} IDE_REGISTERS_2, *PIDE_REGISTERS_2;

/*
 * Identify data without the Reserved4.
 */
typedef struct _IDENTIFY_DATA2 {
	USHORT GeneralConfiguration;            // 00 00
	USHORT NumberOfCylinders;               // 02  1
	USHORT Reserved1;                       // 04  2
	USHORT NumberOfHeads;                   // 06  3
	USHORT UnformattedBytesPerTrack;        // 08  4
	USHORT UnformattedBytesPerSector;       // 0A  5
	USHORT SectorsPerTrack;                 // 0C  6
	USHORT VendorUnique1[3];                // 0E  7-9
	USHORT SerialNumber[10];                // 14  10-19
	USHORT BufferType;                      // 28  20
	USHORT BufferSectorSize;                // 2A  21
	USHORT NumberOfEccBytes;                // 2C  22
	USHORT FirmwareRevision[4];             // 2E  23-26
	USHORT ModelNumber[20];                 // 36  27-46
	UCHAR  MaximumBlockTransfer;            // 5E  47
	UCHAR  VendorUnique2;                   // 5F
	USHORT DoubleWordIo;                    // 60  48
	USHORT Capabilities;                    // 62  49
	USHORT Reserved2;                       // 64  50
	UCHAR  VendorUnique3;                   // 66  51
	UCHAR  PioCycleTimingMode;              // 67
	UCHAR  VendorUnique4;                   // 68  52
	UCHAR  DmaCycleTimingMode;              // 69
	USHORT TranslationFieldsValid:1;        // 6A  53
	USHORT Reserved3:15;
	USHORT NumberOfCurrentCylinders;        // 6C  54
	USHORT NumberOfCurrentHeads;            // 6E  55
	USHORT CurrentSectorsPerTrack;          // 70  56
	ULONG  CurrentSectorCapacity;           // 72  57-58
	USHORT CurrentMultiSectorSetting;       //     59
	ULONG  UserAddressableSectors;          //     60-61
	USHORT SingleWordDMASupport : 8;        //     62
	USHORT SingleWordDMAActive : 8;
	USHORT MultiWordDMASupport : 8;         //     63
	USHORT MultiWordDMAActive : 8;
	USHORT AdvancedPIOModes : 8;            //     64
	USHORT Reserved4 : 8;
	USHORT MinimumMWXferCycleTime;          //     65
	USHORT RecommendedMWXferCycleTime;      //     66
	USHORT MinimumPIOCycleTime;             //     67
	USHORT MinimumPIOCycleTimeIORDY;        //     68
	USHORT Reserved5[2];                    //     69-70
	USHORT ReleaseTimeOverlapped;           //     71
	USHORT ReleaseTimeServiceCommand;       //     72
	USHORT MajorRevision;                   //     73
	USHORT MinorRevision;                   //     74
//    USHORT Reserved6[14];                   //     75-88
} IDENTIFY_DATA2, *PIDENTIFY_DATA2;

#define IDENTIFY_DATA_SIZE sizeof(IDENTIFY_DATA2)

typedef union _CDB {
	/*
	 * Standard 6-byte CDB
	 */
	struct _CDB6READWRITE {
		UCHAR OperationCode;
		UCHAR LogicalBlockMsb1 : 5;
		UCHAR LogicalUnitNumber : 3;
		UCHAR LogicalBlockMsb0;
		UCHAR LogicalBlockLsb;
		UCHAR TransferBlocks;
		UCHAR Control;
	} CDB6READWRITE, *PCDB6READWRITE;

	/*
	 * Standard 10-byte CDB
     */
	struct _CDB10 {
		UCHAR OperationCode;
		UCHAR Reserved1 : 5;
		UCHAR LogicalUnitNumber : 3;
		UCHAR LogicalBlockByte0;
		UCHAR LogicalBlockByte1;
		UCHAR LogicalBlockByte2;
		UCHAR LogicalBlockByte3;
		UCHAR Reserved2;
		UCHAR TransferBlocksMsb;
		UCHAR TransferBlocksLsb;
		UCHAR Control;
	} CDB10, *PCDB10;

	struct _START_STOP {
		UCHAR OperationCode;
		UCHAR Immediate: 1;
		UCHAR Reserved1 : 4;
		UCHAR LogicalUnitNumber : 3;
		UCHAR Reserved2[2];
		UCHAR Start : 1;
		UCHAR LoadEject : 1;
		UCHAR Reserved3 : 6;
		UCHAR Control;
	} START_STOP, *PSTART_STOP;

} CDB, *PCDB;

/*
 * SCSI CDB operation codes
 */
#define SCSIOP_TEST_UNIT_READY     0x00
#define SCSIOP_REQUEST_SENSE       0x03
#define SCSIOP_READ6               0x08
#define SCSIOP_WRITE6              0x0A
#define SCSIOP_INQUIRY             0x12
#define SCSIOP_VERIFY6             0x13
#define SCSIOP_MODE_SENSE          0x1A
#define SCSIOP_START_STOP_UNIT     0x1B
#define SCSIOP_READ_CAPACITY       0x25
#define SCSIOP_READ                0x28
#define SCSIOP_WRITE               0x2A
#define SCSIOP_SEEK                0x2B
#define SCSIOP_VERIFY              0x2F

#ifdef SUPPORT_SATA
#define DRIVER_NAME				"Device Driver for RocketRAID 1520 S-ATA RAID Controller"
#define CONTROLLER_NAME_HPT372A	"RocketRAID 1520 S-ATA Controller"
#else
#define DRIVER_NAME				"Device Driver for HPT3xx ATA RAID Controller"
#define CONTROLLER_NAME_HPT372A	"HPT372A UDMA/ATA133 RAID Controller"
#endif
#define COMPANY_NAME			"HighPoint Technologies, Inc."
#define CONTROLLER_NAME_HPT370	"HPT370 UDMA/ATA100 RAID Controller"
#define CONTROLLER_NAME_HPT372	"HPT372 UDMA/ATA133 RAID Controller"
#define CONTROLLER_NAME_HPT372N	"HPT372N UDMA/ATA133 RAID Controller"
#define CONTROLLER_NAME_HPT302	"HPT302 ATA RAID Controller"
#define CONTROLLER_NAME_HPT302N	"HPT302N ATA RAID Controller"
#define PROC_DIR_NAME			"hpt37x2"
#define HPT_MAX_CMDS			124

typedef struct _PCI1_CFG_ADDR {
	unsigned    reg_num : 8;          // Register number
	unsigned    fun_num : 3;          // function number (0-1)
	unsigned    dev_num : 5;          // device number (0-20)
	unsigned    bus_num : 8;          // bus number (0)
	unsigned    reserved: 7;
	unsigned    enable  : 1;
}  PCI1_CFG_ADDR, *PPCI1_CFG_ADDR;

/*
 * PCI config Register offset
 */
#define REG_VID             0x00    // vender identification register
#define REG_DID             0x02    // device identification register
#define REG_PCICMD          0x04    // command register
#define REG_PCISTS          0x06    // PCI device status register
#define REG_RID             0x08    // revision identification register
#define REG_PI              0x09    // programming interface register
#define REG_SUBC            0x0a    // sub class code register
#define REG_BCC             0x0b    // base class code register
#define REG_MLT             0x0d    // master latency timer register
#define REG_HEDT            0x0e    // header type register
#define REG_IOPORT0         0x10    //
#define REG_IOPORT1         0x14    //
#define REG_BMIBA           0x20    // bus master interface base address reg
#define REG_IRQL			0x3C
#define REG_MISC            0x50    // brooklyn,MISC.control register

#define PCI_IOSEN           0x01    // Enable IO space
#define PCI_BMEN            0x04    // Enable IDE bus master


/*
 * Description:  Scatter-gather table
 */
typedef struct _SCAT_GATH
{
	ULONG  SgAddress;		 	
	USHORT SgSize;              
	USHORT SgFlag;              
}   SCAT_GATH, *PSCAT_GATH;

#define SG_FLAG_EOT         0x8000  
#define MAX_SG_DESCRIPTORS   17   //17 -- maximum 64K

/*
 * HPT Special
 */
#define HPT370_VENDORID 0x1103
#define HPT370_DEVICEID 0x0004
#define HPT372A_VENDORID 0x1103
#define HPT372A_DEVICEID 0x0005
#define HPT302_VENDORID 0x1103
#define HPT302_DEVICEID 0x0006

/*
 * 370-370A timing 
 */
#define PIO_370MODE      0
#define DMA_370MODE      0x2
#define UDMA_370MODE     1

#define DATA_HIGH370(x)  ((ULONG)x)
#define DATA_LOW370(x)   ((ULONG)x << 4)
#define UDMA_CYCLE370(x) ((ULONG)x << 18)
#define DATA_PRE370(x)   ((ULONG)x << 22)
#define CTRL_ENA370(x)   ((ULONG)x << 28)

#define CMD_HIGH370(x)   ((ULONG)x << 9)
#define CMD_LOW370(x)    ((ULONG)x << 13)
#define CMD_PRE370(x)    ((ULONG)x << 25)

#define CLK50_PIO370     (CTRL_ENA370(PIO_370MODE)|CMD_PRE370(5)|CMD_LOW370(15)|CMD_HIGH370(10))

#define CLK50_370PIO0    (CLK50_PIO370|DATA_PRE370(3)|DATA_LOW370(8)|DATA_HIGH370(10))
#define CLK50_370PIO1    (CLK50_PIO370|DATA_PRE370(3)|DATA_LOW370(6)|DATA_HIGH370(5))
#define CLK50_370PIO2    (CLK50_PIO370|DATA_PRE370(2)|DATA_LOW370(5)|DATA_HIGH370(4))
#define CLK50_370PIO3    (CLK50_PIO370|DATA_PRE370(2)|DATA_LOW370(4)|DATA_HIGH370(3))
#define CLK50_370PIO4    (CLK50_PIO370|DATA_PRE370(2)|DATA_LOW370(4)|DATA_HIGH370(2))

#define CLK50_DMA370     (CTRL_ENA370(DMA_370MODE)|CMD_PRE370(1)|CMD_LOW370(4)|CMD_HIGH370(1))
#define CLK50_370DMA0    (CLK50_DMA370|DATA_PRE370(2)|DATA_LOW370(14)|DATA_HIGH370(10))
#define CLK50_370DMA1    (CLK50_DMA370|DATA_PRE370(2)|DATA_LOW370(5)|DATA_HIGH370(4))
#define CLK50_370DMA2    (CLK50_DMA370|DATA_PRE370(2)|DATA_LOW370(4)|DATA_HIGH370(2))

#define CLK50_UDMA370    (CTRL_ENA370(UDMA_370MODE)|CMD_PRE370(1)|CMD_LOW370(4)|CMD_HIGH370(1))
#define CLK50_370UDMA0   (CLK50_UDMA370|UDMA_CYCLE370(6)|DATA_LOW370(14)|DATA_HIGH370(10))
#define CLK50_370UDMA1   (CLK50_UDMA370|UDMA_CYCLE370(5)|DATA_LOW370(5)|DATA_HIGH370(4))
#define CLK50_370UDMA2   (CLK50_UDMA370|UDMA_CYCLE370(3)|DATA_LOW370(4)|DATA_HIGH370(2))
#define CLK50_370UDMA3   (CLK50_UDMA370|UDMA_CYCLE370(3)|DATA_PRE370(2)|DATA_LOW370(4)|DATA_HIGH370(2))
#define CLK50_370UDMA4   (CLK50_UDMA370|UDMA_CYCLE370(11)|DATA_PRE370(2)|DATA_LOW370(4)|DATA_HIGH370(2))
#define CLK50_370UDMA5   (CLK50_UDMA370|UDMA_CYCLE370(1)|DATA_PRE370(2)|DATA_LOW370(4)|DATA_HIGH370(2))

#define CLK33_PIOCMD370  (CTRL_ENA370(PIO_370MODE)|CMD_PRE370(3)|CMD_LOW370(10)|CMD_HIGH370(7))
#define CLK33_370PIO0    (CLK33_PIOCMD370|DATA_PRE370(2)|DATA_LOW370(10)|DATA_HIGH370(7))
#define CLK33_370PIO1    (CLK33_PIOCMD370|DATA_PRE370(2)|DATA_LOW370(9)|DATA_HIGH370(3))
#define CLK33_370PIO2    (CLK33_PIOCMD370|DATA_PRE370(1)|DATA_LOW370(5)|DATA_HIGH370(3))
#define CLK33_370PIO3    (CLK33_PIOCMD370|DATA_PRE370(1)|DATA_LOW370(4)|DATA_HIGH370(2))
#define CLK33_370PIO4    (CLK33_PIOCMD370|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(1))

#define CLK33_DMACMD370  (CTRL_ENA370(DMA_370MODE)|CMD_PRE370(1)|CMD_LOW370(3)|CMD_HIGH370(6))
#define CLK33_370DMA0    (CLK33_DMACMD370|DATA_PRE370(1)|DATA_LOW370(9)|DATA_HIGH370(7))
#define CLK33_370DMA1    (CLK33_DMACMD370|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(3))
#define CLK33_370DMA2    (CLK33_DMACMD370|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(1))

#define CLK33_UDMACMD370 (CTRL_ENA370(UDMA_370MODE)|CMD_PRE370(1)|CMD_LOW370(3)|CMD_HIGH370(1))
#define CLK33_370UDMA0   (CLK33_UDMACMD370|UDMA_CYCLE370(4)|DATA_PRE370(1)|DATA_LOW370(9)|DATA_HIGH370(7))
#define CLK33_370UDMA1   (CLK33_UDMACMD370|UDMA_CYCLE370(3)|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(3))
#define CLK33_370UDMA2   (CLK33_UDMACMD370|UDMA_CYCLE370(2)|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(1))
#define CLK33_370UDMA3   (CLK33_UDMACMD370|UDMA_CYCLE370(11)|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(1))
#define CLK33_370UDMA4   (CLK33_UDMACMD370|UDMA_CYCLE370(1)|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(1))
#define CLK33_370UDMA5   (CLK33_UDMACMD370|UDMA_CYCLE370(1)|DATA_PRE370(1)|DATA_LOW370(3)|DATA_HIGH370(1))

/*
 * io commands
 */
#define IO_COMMAND_READ		1
#define IO_COMMAND_WRITE	2

/************************************************************************
 * shared data structures
 ************************************************************************/
/*
 * Chip Type
 */ 
#define CHIP_TYPE_HPT366      1
#define CHIP_TYPE_HPT368      2
#define CHIP_TYPE_HPT370      3
#define CHIP_TYPE_HPT370A     4
#define CHIP_TYPE_HPT370B     5
#define CHIP_TYPE_HPT374      6
#define CHIP_TYPE_HPT372      7
#define CHIP_TYPE_HPT372A     8
#define CHIP_TYPE_HPT302      9
#define CHIP_TYPE_HPT371      10
#define CHIP_TYPE_HPT372N     11
#define CHIP_TYPE_HPT302N     12

/*
 * ioctl structure
 */
#define HPT_IOCTL_MAGIC 0x1A2B3C4D

typedef struct _HPT_IOCTL_PARAM {
	DWORD Magic;				
	DWORD dwIoControlCode;       
	PVOID lpInBuffer;            
	DWORD nInBufferSize;      
	PVOID lpOutBuffer;          
	DWORD nOutBufferSize;    
	PDWORD lpBytesReturned; 
}
HPT_IOCTL_PARAM, *PHPT_IOCTL_PARAM;  	

/* 
 * ioctl command
 */
#define HPT_IOCTL_CMD	0xFE

/*
 * SRB Functions
 */
#define SRB_FUNCTION_EXECUTE_SCSI           0x00
#define SRB_FUNCTION_IO_CONTROL             0x02
#define SRB_FUNCTION_SHUTDOWN               0x07
#define SRB_FUNCTION_FLUSH                  0x08
/*
 * SRB Status
 */
#define SRB_STATUS_PENDING                  0x00
#define SRB_STATUS_SUCCESS                  0x01
#define SRB_STATUS_ABORTED                  0x02
#define SRB_STATUS_ERROR                    0x04
#define SRB_STATUS_BUSY                     0x05
#define SRB_STATUS_INVALID_REQUEST          0x06
#define SRB_STATUS_INVALID_PATH_ID          0x07
#define SRB_STATUS_NO_DEVICE                0x08
#define SRB_STATUS_SELECTION_TIMEOUT        0x0A
#define SRB_STATUS_BUS_RESET                0x0E
#define SRB_STATUS_NO_HBA                   0x11
#define SRB_STATUS_INVALID_LUN              0x20
#define SRB_STATUS_INVALID_TARGET_ID        0x21
#define SRB_STATUS_NEED_REQUEUE             0x24
/*
 * SRB Status Masks
 */
#define SRB_STATUS_QUEUE_FROZEN             0x40
#define SRB_STATUS_AUTOSENSE_VALID          0x80

#define SRB_STATUS(Status) (Status & ~(SRB_STATUS_AUTOSENSE_VALID | SRB_STATUS_QUEUE_FROZEN))

/*
 * SRB Flag Bits
 */
#define SRB_FLAGS_DATA_IN                   0x00000040
#define SRB_FLAGS_DATA_OUT                  0x00000080

typedef struct _SrbExtension {
	ULONG      StartLBA;   

    UCHAR      FirstMember;    
    UCHAR      LastMember;     
    UCHAR      InSameLine;     
    UCHAR      reserve1;

    USHORT     FirstSectors;   
    USHORT     FirstOffset;    

    USHORT     LastSectors;    
    USHORT     AllMemberBlocks;

	SCAT_GATH  ArraySg[MAX_SG_DESCRIPTORS]; 

	ADDRESS    DataBuffer;    
	USHORT     DataTransferLength;
	USHORT     SgFlags;    

	USHORT     JoinMembers;    
	USHORT     WaitInterrupt;  

	ULONG      Lba;            
	USHORT     nSector;        
	UCHAR      SrbFlags;       

	struct _Device *pMaster;   

	ULONG       	WorkingFlags;

	Scsi_Cmnd       *pREQ; 				
	struct _Device  *pDev;				

} SrbExtension, *PSrbExtension;

/* SRB Working flags define area */
#define	SRB_WFLAGS_USE_INTERNAL_BUFFER		0x00000001 
#define	SRB_WFLAGS_IGNORE_ARRAY				0x00000002 
#define	SRB_WFLAGS_HAS_CALL_BACK			0x00000004 
#define	SRB_WFLAGS_MUST_DONE				0x00000008 
#define	SRB_WFLAGS_ON_MIRROR_DISK			0x00000010 
#define	SRB_WFLAGS_ON_SOURCE_DISK			0x00000020 
#define SRB_WFLAGS_ARRAY_IO_STARTED         0x10000000 
#define SRB_WFLAGS_WATCHTIMER_CALLED		0x20000000 
#define SRB_WFLAGS_USE_SG					0x40000000 

/*
 * SCSI I/O Request Block
 */
typedef struct _SCSI_REQUEST_BLOCK {
	USHORT Length;
	UCHAR Function;
	UCHAR SrbStatus;
	UCHAR ScsiStatus;
	UCHAR CdbLength;
	UCHAR SenseInfoBufferLength;
	UCHAR reseved;
	ULONG SrbFlags;
	ULONG DataTransferLength;
	ULONG TimeOutValue;
	PVOID DataBuffer;
	PVOID SenseInfoBuffer;
	PVOID NextSrb;
	UCHAR Cdb[16];
	/* buffer to hold SrbExtension
	 * it must use 16-byte align.
	 */
	PVOID SrbExtension;
	UCHAR _SrbExtensionBuffer[sizeof(SrbExtension) + 16];
} SCSI_REQUEST_BLOCK, *PSCSI_REQUEST_BLOCK;

#define SCSI_REQUEST_BLOCK_SIZE sizeof(SCSI_REQUEST_BLOCK)

struct _Channel;  
/*
 * Description: _Device
 */
typedef struct _Device {
	UINT    DeviceFlags;       
	UINT	DeviceFlags2;	   
	UCHAR   UnitId;            
	UCHAR   DeviceModeSetting; 
					
	UCHAR   ReadCmd;           
	UCHAR   WriteCmd;          
	UINT    MultiBlockSize;    

	USHORT  RealHeadXsect;     
	UCHAR   RealSector;        
	UCHAR   RealHeader;        

	UCHAR   bestPIO;           
	UCHAR   bestDMA;           
	UCHAR   bestUDMA;          
	UCHAR   Usable_Mode;       

	ULONG   HidenLBA;          
	struct _Channel *pChannel; 

	ULONG   capacity;          

	IDENTIFY_DATA2 IdentifyData;

	UCHAR   SmartCommand;      
	UCHAR   ReturningMediaStatus; 

	UCHAR   ArrayNum;		
	UCHAR   reserved1;
	USHORT  ArrayMask;         
	USHORT  reserved2;
	struct _VirtualDevice *pArray;

	struct{
		ULONG	nLastError;				
		struct _Device	*pNextErrorDevice;		
		UCHAR	Cdb[16];
	} stErrorLog;

	ULONG   RebuiltSector;

	UINT   IoCount; 
	UINT   ResetCount; 
	UINT   IoSuccess;	

} Device, *PDevice;

/*
 *  First device flags declare
 */
#define DFLAGS_LBA                  0x1    
#define DFLAGS_DMA                  0x2    
#define DFLAGS_MULTIPLE             0x4    
#define DFLAGS_ULTRA                0x8    

#define DFLAGS_DMAING               0x10   
#define DFLAGS_REQUEST_DMA          0x20   
#define DFLAGS_HIDEN_DISK           0x40   
#define DFLAGS_ARRAY_DISK           0x80   

#define DFLAGS_NEED_SWITCH          0x200  
#define DFLAGS_REMOVABLE_DRIVE      0x400  

#define DFLAGS_CDROM_DEVICE         0x1000 
#define DFLAGS_HARDDISK             0x2000 
#define DFLAGS_LS120                0x4000 
#define DFLAGS_BOOT_SECTOR_PROTECT  0x8000 

#define DFLAGS_48BIT_LBA         	0x8000000

#define DFLAGS_TAPE_RDP             0x20000 
#define DFLAGS_MEDIA_STATUS_ENABLED 0x40000 
#define DFLAGS_TAPE_DEVICE          0x80000 

#define DFLAGS_SET_CALL_BACK        0x100000
#define DFLAGS_SORT_UP              0x200000
#define DFLAGS_ARRAY_WAIT_EXEC      0x400000

#define DFLAGS_WIN_SHUTDOWN			0x1000000
#define DFLAGS_WIN_FLUSH   			0x2000000

/*Second device flags*/
#define DFLAGS_REDUCE_MODE	        0x00010000
#define DFLAGS_DEVICE_DISABLED	    0x00020000	
#define DFLAGS_BOOTABLE_DEVICE		0x00080000	
#define DFLAGS_BOOT_MARK			0x00100000  
#define DFLAGS_NEW_ADDED			0x40000000  

#define DFLAGS_REMAINED_MEMBER		0x80000000

#define ARRAY_BROKEN    0xFF   /* some members lost in the array */
#define DEVICE_REMOVED  0xFE   /* some members was removed by hot plug */
#define DEVICE_PLUGGED	0xFD   /* some members was added by hot plug */

struct fdisk_partition_table {
  unsigned char bootid;   /* bootable?  0=no, 128=yes  */
  unsigned char beghead;  /* beginning head number */
  unsigned char begsect;  /* beginning sector number */
  unsigned char begcyl;   /* 10 bit nmbr, with high 2 bits put in begsect */
  unsigned char systid;   /* Operating System type indicator code */
  unsigned char endhead;  /* ending head number */
  unsigned char endsect;  /* ending sector number */
  unsigned char endcyl;   /* also a 10 bit nmbr, with same high 2 bit trick */
  int relsect;            /* first sector relative to start of disk */
  int numsect;            /* number of sectors in partition */
};

struct master_boot_record {
  char    bootinst[446];   /* space to hold actual boot code */
  struct fdisk_partition_table  parts[4];
  unsigned short  signature;       /* set to 0xAA55 to indicate PC MBR format */
};

/*
 * Description: _Channel
 */
typedef struct _Channel {
	PIDE_REGISTERS_1 BaseIoAddress1;  
	PIDE_REGISTERS_2 BaseIoAddress2;  
	PUCHAR           BMI;             
	PUCHAR           BaseBMI;         

	PUCHAR			 NextChannelBMI;  
	PUCHAR			 MiscControlAddr; 

	UCHAR            InterruptLevel;
	UCHAR            reserve;
	USHORT           ChannelFlags;    

	char	         adjusted;    
	UCHAR            RetryTimes;      
	USHORT           nSector;         
	ULONG            Lba;             

	ADDRESS          BufferPtr;       
	UINT             WordsLeft;       

	UINT             exclude_index;   

	PDevice          pWorkDev;        

	PCI1_CFG_ADDR    pci1_cfg;        
	ULONG            *Setting;        

	PDevice          pDevice[2];      
	Device           Devices[2];

	ULONG            SgPhysicalAddr;  
	PSCAT_GATH       pSgTable;        

	UCHAR            SGBuf[MAX_SG_DESCRIPTORS * sizeof(SCAT_GATH) + 16];

	PSCSI_REQUEST_BLOCK CurrentSrb;   
	SCSI_REQUEST_BLOCK  _Srb;         

	ULONG			 RetryCount;     

	struct timer_list *retry_timer;
	struct timer_list *watch_timer;
	struct timer_list *busy_timer;

} Channel, *PChannel;

typedef struct _Adapter {
	char *  name;
	UCHAR   chip_type;
	UCHAR   num_buses;
	UCHAR	irq;
	UCHAR	irqOwned;
	UCHAR   pci_bus;
	UCHAR	devfn;
	UCHAR	offline;
	UCHAR   reserved2;
	Channel *IDEChannel;
}
HPT_ADAPTER, *PHPT_ADAPTER;

/*
 * Description: Virtual Device Table
 */
typedef struct _VirtualDevice {
    UCHAR   nDisk;             
    UCHAR   BlockSizeShift;    
    UCHAR   arrayType;         
    UCHAR   pad_1;
    
    WORD    ArrayNumBlock;     
    WORD    pad_2;

    ULONG   Stamp;             
	ULONG	MirrorStamp;	

    ULONG   capacity;          

    PSCSI_REQUEST_BLOCK Srb;   

	ULONG	RaidFlags;	

	struct _Device  *pDevice[MAX_MEMBERS]; 

	UCHAR				ArrayName[32];	

	ULONG   RebuiltSector;	
	
	ULONG   LockedLba;      
	ULONG   LockedSectors;  

} VirtualDevice, *PVirtualDevice;

/*
 * RAID FLAGS declare area
 */
#define RAID_FLAGS_NEED_SYNCHRONIZE		0x00000001
#define RAID_FLAGS_NEED_REBUILD 		RAID_FLAGS_NEED_SYNCHRONIZE
#define RAID_FLAGS_BROKEN               0x00000002
#define RAID_FLAGS_BEING_BUILT			0x00000004
#define RAID_FLAGS_DISABLED				0x00000008
#define RAID_FLAGS_BOOTDISK				0x00000010
#define RAID_FLAGS_NEWLY_CREATED		0x00000020
#define RAID_FLAGS_ABORT_REBUILD		0x00000040
#define RAID_FLAGS_DUPLICATE_AND_CREATE 0x00000080

#define RAID_FLAG_SET(pArray, f) (pArray)->RaidFlags |= (f)
#define RAID_FLAG_CLEAR(pArray, f) (pArray)->RaidFlags &= ~(f)
#define RAID_FLAG_ISSET(pArray, f) (((pArray)->RaidFlags & (f)) !=0)

#define MIRROR_DISK    (MAX_MEMBERS - 1)

/* arrayType */
#define VD_RAID_0_STRIPE     0 
#define VD_RAID_1_MIRROR     1 
#define VD_RAID_01_2STRIPE   2 
#define VD_SPAN              3 
#define VD_RAID01_MIRROR     7 
#define VD_INVALID_TYPE		0xFF

#define SPARE_DISK     (MAX_MEMBERS - 2)

extern UINT  hpt_NumAdapters;                 /* Number of Adapters */
#define NumAdapters hpt_NumAdapters
extern PHPT_ADAPTER hpt_adapters[];
extern struct Scsi_Host *hpt_vhost; /* virtual scsi host */

extern VirtualDevice VirtualDevices[MAX_V_DEVICE];
extern PVirtualDevice pLastVD;

typedef struct _logical_device {
	UCHAR   isArray;
	UCHAR   isValid;
	UCHAR	isInUse;
	UCHAR   reserved;
	PVOID   pLD;
}
LOGICAL_DEVICE;

extern LOGICAL_DEVICE hpt_logical_devices[];
extern DWORD hpt_num_logical_devices;
extern ULONG hpt_page_offset;

typedef struct _hpt_GET_INFO {
	char *buffer;
	int buflength;
	int bufoffset;
	int buffillen;
	int filpos;	
} HPT_GET_INFO;

#define MAX_DPC 16
typedef struct st_HPT_DPC {
	void (*dpc)(void *);
	void *arg;
} ST_HPT_DPC;

extern volatile int stop_thread;
extern struct semaphore * hpt_notify_wt;
extern struct semaphore * hpt_queue_ready;
typedef void (*HPT_DPC)(void*);

extern ST_HPT_DPC FastDpcQueue[MAX_DPC];
extern volatile int FastDpcQueue_First;
extern volatile int FastDpcQueue_Last;

/*
 * time represented in DWORD format
 */
typedef struct _TIME_RECORD {
   UINT        seconds:6;      /* 0 - 59 */
   UINT        minutes:6;      /* 0 - 59 */
   UINT        month:4;        /* 1 - 12 */
   UINT        hours:6;        /* 0 - 59 */
   UINT        day:5;          /* 1 - 31 */
   UINT        year:5;         /* 0=2000, 31=2031 */
} TIME_RECORD;

typedef DWORD DEVICEID;

/* 
 * event structure
 */
typedef struct _HPT_EVENT {
	TIME_RECORD Time;
	DEVICEID    DeviceID;
	UCHAR       EventType;
	UCHAR		reserved[3];

	UCHAR		Data[32]; /* various data depend on EventType */
} HPT_EVENT, *PHPT_EVENT;

/* ChannelFlags */
#define IS_HPT_370      1      /* This is a HPT370 Adapter */
#define IS_HPT_372      2      /* This is a HPT370 Adapter */
#define IS_HPT_372A     4      /* This is a HPT372A Adapter */
#define IS_DPLL_MODE    0x8    /* HPT370 is using DPLL */
#define IS_80PIN_CABLE  0x30   /* 80 pin cable attached */
#define PF_ACPI_INTR    0x80   /* hptpwr.c is waiting a interrupt */
#define IS_HPT_372N     0x100
#define IS_HPT_302N     0x200
#define IS_HPT_302		0x400


/*
 * Global Data
 */
extern UINT exlude_num;
extern ULONG excluded_flags;

#define MAX_EVENTS 10
extern HPT_EVENT hptdev_event_queue[];
volatile extern int event_queue_head, event_queue_tail;


/* Macros */
#define KMSG_LEADING PROC_DIR_NAME": "
#define hpt_printk(_x_) do { printk(KMSG_LEADING); printk _x_ ; } while (0)

/* DebugPrint */
#ifdef DEBUG
extern int hpt_dbg_level;
#define KdPrintE(_x_) do { if (hpt_dbg_level>0) { printk(KMSG_LEADING); printk _x_ ;} } while (0)
#define KdPrintW(_x_) do { if (hpt_dbg_level>1) { printk(KMSG_LEADING); printk _x_ ;} } while (0)
#define KdPrintI(_x_) do { if (hpt_dbg_level>2) { printk(KMSG_LEADING); printk _x_ ;} } while (0)
#else
#define KdPrintI(_x_)
#define KdPrintW(_x_)
#define KdPrintE(_x_)
#endif // DEBUG
#define KdPrint(_x_) KdPrintI(_x_)

#define MapToLogicAddr(x)            ((void *)(hpt_page_offset | (unsigned long)(x)))
#define MapToPhysAddr(x)             ((unsigned long)(x) & ~hpt_page_offset)

#define MAX_RETRY_COUNT 5
#define ScheduleRetryProcess(pChan) do {\
			pChan->retry_timer->expires = jiffies + 10; \
			add_timer(pChan->retry_timer);\
		} while (0)
#define CancelRetryProcess(pChan) del_timer(pChan->retry_timer)

/*
 * Busy-handle Timer, Hz=ticks/second of the clock
 */
#define SetBusyTimer(pChan) do {\
			pChan->busy_timer->expires = jiffies + (HZ>>10) + 1; \
			add_timer(pChan->busy_timer);\
		} while (0)
#define CancelBusyTimer(pChan) del_timer(pChan->busy_timer)

/* Minimum and maximum macros */
#define MAX(a, b)	(((a) > (b)) ? (a) : (b))
#define MIN(a, b)	(((a) < (b)) ? (a) : (b))
#define shortswap(w) ((WORD)((w)>>8 | ((w) & 0xFF)<<8))

/* Operating System macro */
/* gmm 2001-5-29
 *  If we use virt_to_bus and bus_to_virt, we have to take care of
 *  different PAGE_OFFSET of the running kernel.
 */
#define GetCurrentTime()             (jiffies)
#define T_IOPORT(x) ((USHORT)(int)(x))
#define OutDWord(x, y)    outl(y, T_IOPORT(x))
#define InDWord(x)        inl(T_IOPORT(x))
#define OutPort(x, y)     outb(y, T_IOPORT(x))
#define InPort(x)         inb(T_IOPORT(x))
#define OutWord(x, y)     outw(y, T_IOPORT(x))
#define InWord(x)         inw(T_IOPORT(x))
#define RepINS(x,y,z)     insw(T_IOPORT(&x->Data), (PUSHORT)y, z)
#define RepOUTS(x,y,z)    outsw(T_IOPORT(&x->Data), (PUSHORT)y, z)
#define RepINSD(x,y,z)    insd(T_IOPORT(&x->Data), (PULONG)y, (z) >> 1)
#define RepOUTSD(x,y,z)   outsd(T_IOPORT(&x->Data), (PULONG)y, (z) >> 1)
#define OS_RepINS(x,y,z)  RepINS(x,y,z)

#define GetStatus(IOPort2)           (UCHAR)InPort(&IOPort2->AlternateStatus)
#define UnitControl(IOPort2, Value)  OutPort(&IOPort2->AlternateStatus,(UCHAR)(Value))

#define GetErrorCode(IOPort)         (UCHAR)InPort((PUCHAR)&IOPort->Data+1)
#define SetFeaturePort(IOPort,x)     OutPort((PUCHAR)&IOPort->Data+1, x)
#define SetBlockCount(IOPort,x)      OutPort(&IOPort->BlockCount, x)
#define GetInterruptReason(IOPort)   (UCHAR)InPort(&IOPort->BlockCount)
#define SetBlockNumber(IOPort,x)     OutPort(&IOPort->BlockNumber, x)
#define GetBlockNumber(IOPort)       (UCHAR)InPort((PUCHAR)&IOPort->BlockNumber)
#define GetByteLow(IOPort)           (UCHAR)InPort(&IOPort->CylinderLow)
#define SetCylinderLow(IOPort,x)     OutPort(&IOPort->CylinderLow, x)
#define GetByteHigh(IOPort)          (UCHAR)InPort(&IOPort->CylinderHigh)
#define SetCylinderHigh(IOPort,x)    OutPort(&IOPort->CylinderHigh, x)
#define GetBaseStatus(IOPort)        (UCHAR)InPort(&IOPort->Command)

#define SelectUnit(IOPort,UnitId) \
	do { \
		OutPort(&IOPort->DriveSelect, (UCHAR)(UnitId));\
		OutPort(&IOPort->DriveSelect, (UCHAR)(UnitId));\
	} while (0)
#define GetCurrentSelectedUnit(IOPort) (UCHAR)InPort(&IOPort->DriveSelect)

#define IssueCommand(IOPort,Cmd)     OutPort(&IOPort->Command, (UCHAR)Cmd)

#define DisableBoardInterrupt(x) OutPort(x+0x7A, 0x10)
#define EnableBoardInterrupt(x)	OutPort(x+0x7A, 0x0)

#define SetMaxCmdQueue(x, y)
#define OS_Identify(x)  memcpy((char *)&pDev->IdentifyData, (char *)&Identify, \
						sizeof(IDENTIFY_DATA2))

#define MemoryCopy(x,y,z) memcpy((char *)(x), (char *)(y), z)
#define GetStamp()		GetCurrentTime()
#define CopyInternalBuffer(pDev, Srb)
#define OS_Set_Array_Wait_Flag(pDevice) pDevice->DeviceFlags |= DFLAGS_ARRAY_WAIT_EXEC

#ifndef NOT_ISSUE_37
	#ifdef  ISSUE_37_ONLY
		#define Reset370IdeEngine(pChan, UnitId) do {\
					if(pChan->ChannelFlags & IS_HPT_370) \
						 OutPort(pChan->BMI + (((UINT)pChan->BMI & 0xF)? 0x6C : 0x70), 0x37); \
					} while (0)
	#else //ISSUE_37_ONLY
		void INLINE Reset370IdeEngine(PChannel pChan, UCHAR UnitId);
	#endif //ISSUE_37_ONLY
	void INLINE Switching370Clock(PChannel pChan, UCHAR ucClockValue);
#endif //NOT_ISSUE_37

/*
 * Function prototype
 */
UCHAR DeviceInterrupt(PDevice pDev, UINT flags);
#define ZeroMemory(p, size)			memset(p, 0, size)
void get_array_name(char *arrayname, PVirtualDevice pArray);
void SetHptChip(PChannel Primary, PUCHAR BMI);
int hpt_queue_dpc(HPT_DPC dpc, void *arg);
#ifdef DRIVER_REBUILD
void hpt_rebuild_mirror(PVirtualDevice pArray);
#endif
void StallExec(ULONG x);
int CopySgl(PSCAT_GATH pSG, Scsi_Cmnd *pREQ, int isphy);
BOOLEAN AtaPioInterrupt(PDevice pDevice);
void NewIdeIoCommand(PDevice pDevice , PSCSI_REQUEST_BLOCK Srb);
int btr(ULONG locate);
void bts(ULONG locate);
PDevice ReportError(PDevice pDev, BYTE error, PSCSI_REQUEST_BLOCK Srb);
void hpt_rescan_all_devices(void);
void check_bootable(PDevice pDevice);
void hpt_check_plugged_device(PDevice pDev);
int  IdeHardReset(PIDE_REGISTERS_1 IoPort, PIDE_REGISTERS_2 ControlPort);
void IdeMediaStatus(BOOLEAN EnableMSN, IN PDevice pDev);
BOOL device_rdwr(PDevice pDev, int cmd, ULONG lba, UCHAR nSector, PVOID buffer, DWORD flags);
int ReadWrite(PDevice pDev, ULONG Lba, UCHAR Cmd , PUSHORT tmpBuffer);
void DeviceSelectMode(PDevice pDev, UCHAR NewMode);
void IdeResetController(PChannel pChan);
int BuildSgl(PDevice pDev, PSCAT_GATH p , PSCSI_REQUEST_BLOCK Srb);
VOID TaskDone(PSCSI_REQUEST_BLOCK Srb);
VOID TaskQueue(void);

void Prepare_Shutdown(PDevice pDev);
void StartIdeCommand(PDevice pDevice, PSCSI_REQUEST_BLOCK Srb);
BOOL Device_IsRemoved(PDevice pDev);
int hpt_queue_fastdpc(HPT_DPC dpc, void *arg);
UCHAR WaitOnBusy(PIDE_REGISTERS_2 BaseIoAddress) ;
void hpt_refresh_logical_devices(void);
int event_queue_add(PHPT_EVENT pEvent);
int event_queue_remove(PHPT_EVENT pEvent);
DEVICEID MAKE_DEVICE_ID(PDevice pDev);
void hpt_set_device_on_offline(int id, int online);

#ifdef SUPPORT_ARRAY
BOOLEAN CheckNextRequest(PChannel pChan);
void disk_plugged_dpc(PDevice pDev);
void CheckArray(IN PDevice pDevice);
BOOLEAN AtapiResetController(PChannel pChan);
BOOLEAN ArrayInterrupt(PDevice pDev);
PDevice ReportArray(PDevice pDev, BYTE error, PSCSI_REQUEST_BLOCK Srb);
void R01_member_fail(PDevice pDev);
void Span_SG_Table(PDevice pDevice, PSCAT_GATH p , PSrbExtension  pSrbExt);
void Span_Lba_Sectors(PDevice pDevice , PSrbExtension  pSrbExt);
void Stripe_SG_Table(PDevice pDevice, PSCAT_GATH p , PSrbExtension  pSrbExt);
void Stripe_Lba_Sectors(PDevice pDevice , PSrbExtension  pSrbExt);
void Final_Array_Check(void);
int hpt_set_asc_array_info(char *buffer, int length);
int Kernel_DeviceIoControl(
							DWORD dwIoControlCode,       // operation control code
							PVOID lpInBuffer,            // input data buffer
							DWORD nInBufferSize,         // size of input data buffer
							PVOID lpOutBuffer,           // output data buffer
							DWORD nOutBufferSize,        // size of output data buffer
							PDWORD lpBytesReturned      // byte count
						);
void hpt_set_remained_member(PDevice pDev);
#ifdef SUPPORT_IOCTL
extern void ioctl_ReportEvent(UCHAR event, PVOID param);
#define REPORT_EVENT(e, p) ioctl_ReportEvent(e, p)
#else
#define REPORT_EVENT(e, p)
#endif
#endif

/*
 * kernel function wrapper
 */
extern __attribute__((regparm(0))) int printk_Rhpt(const char * fmt, ...) __attribute__ ((format (printf, 1, 2)));

extern void outb_Rhpt(unsigned char value, unsigned short port);
extern unsigned long volatile hpt_get_jiffies(void);
extern int hpt_get_HZ(void);
extern void *kmalloc_Rhpt(unsigned int size,int priority);
extern void kfree_Rhpt(const void *objp);
extern void schedule_Rhpt(void);

#pragma pack()
#endif //_HPTGLB_H_
